{-# LANGUAGE OverloadedStrings #-}

module Pipes.IRC.Server.IrcMonad
where

import           Control.Lens
import           Control.Monad.RWS
import           Data.ByteString.Char8   (ByteString)
import qualified Data.ByteString.Char8   as BS
import qualified Data.Map                as M
import           Data.Maybe              (catMaybes, fromJust, fromMaybe,
                                          isJust, isNothing)
import qualified Data.Set                as S
import           Pipes.IRC.Message.Types
import           Pipes.IRC.Server.Types


-- | IrcUser management

mkUser :: IrcMonad IrcUser
mkUser = do
  conn <- use clientConn
  srvname <- view ircHostName
  return $ newUser srvname conn

newUser :: ByteString -> Int -> IrcUser
newUser srvname cid =
  IrcUser { _userServerName = srvname
          , _userModes = S.empty
          , _userChannels = S.empty
          , _userConn = cid
          , _userInvites = S.empty
          }

userAddChan, userDelChan :: ByteString -> IrcUser -> IrcUser
userAddChan cn = userChannels %~ S.insert cn
userDelChan cn = userChannels %~ S.delete cn

userAddMode, userDelMode :: IrcUserMode -> IrcUser -> IrcUser
userAddMode um = userModes %~ S.insert um
userDelMode um = userModes %~ S.delete um

userAddInvite, userDelInvite :: ByteString -> IrcUser -> IrcUser
userAddInvite cn = userInvites %~ S.insert cn
userDelInvite cn = userInvites %~ S.delete cn

userHasMode :: IrcUserMode -> IrcUser -> Bool
userHasMode um usr = S.member um $ usr ^. userModes

userInChan :: ByteString -> IrcUser -> Bool
userInChan cn usr = S.member cn $ usr ^. userChannels

-- | IrcChannel management

newChannel :: ByteString -> IrcChannel
newChannel creator = IrcChannel { _chanTopic = Nothing
                                , _chanKey   = Nothing
                                , _chanModeFlags = S.empty
                                , _chanUsers = S.fromList [creator]
                                , _chanOpers = S.fromList [creator]
                                , _chanVoices = S.empty
                                , _chanInvites = S.empty
                                }

chanAddModeFlag, chanDelModeFlag :: IrcChanModeFlags -> IrcChannel -> IrcChannel
chanAddModeFlag cm = chanModeFlags %~ S.insert cm
chanDelModeFlag cm = chanModeFlags %~ S.delete cm

chanAddUser, chanDelUser :: ByteString -> IrcChannel -> IrcChannel
chanAddUser un = chanUsers %~ S.insert un
chanDelUser un = (chanUsers %~ S.delete un)
               . (chanOpers %~ S.delete un)
               . (chanVoices %~ S.delete un)

chanHasUser :: ByteString -> IrcChannel -> Bool
chanHasUser un ch = S.member un $ ch ^. chanUsers

chanSetTopic :: ByteString -> IrcChannel -> IrcChannel
chanSetTopic top = chanTopic .~ Just top

chanHasTopic :: IrcChannel -> Bool
chanHasTopic ch = isJust $ ch ^. chanTopic

chanHasModeFlag :: IrcChanModeFlags -> IrcChannel -> Bool
chanHasModeFlag cm ch = S.member cm $ ch ^. chanModeFlags

chanSigil :: IrcChannel -> ByteString
chanSigil ch | chanHasModeFlag Secret ch  = "@"
             | chanHasModeFlag Private ch = "*"
             | otherwise                  = "="

chanUserSigil :: ByteString -> IrcChannel -> ByteString
chanUserSigil un ch | S.member un $ ch ^. chanOpers  = "@"
                    | S.member un $ ch ^. chanVoices = "+"
                    | otherwise                      = ""

chanSetKey :: ByteString -> IrcChannel -> IrcChannel
chanSetKey key = chanKey .~ Just key

chanHasKey :: IrcChannel -> Bool
chanHasKey ch = isJust $ ch ^. chanKey

chanCheckKey :: ByteString -> IrcChannel -> Bool
chanCheckKey key ch = case ch ^. chanKey of
  Just chKey -> key == chKey
  Nothing    -> True

chanAddOper, chanDelOper :: ByteString -> IrcChannel -> IrcChannel
chanAddOper un = chanOpers %~ S.insert un
chanDelOper un = chanOpers %~ S.delete un

chanAddVoice, chanDelVoice :: ByteString -> IrcChannel -> IrcChannel
chanAddVoice un = chanVoices %~ S.insert un
chanDelVoice un = chanVoices %~ S.delete un

chanAddInvite, chanDelInvite :: ByteString -> IrcChannel -> IrcChannel
chanAddInvite un = chanInvites %~ S.insert un
chanDelInvite un = chanInvites %~ S.delete un

chanUserIsOper :: ByteString -> IrcChannel -> Bool
chanUserIsOper un ch = S.member un $ ch ^. chanOpers

chanUserHasVoice :: ByteString -> IrcChannel -> Bool
chanUserHasVoice un ch = S.member un $ ch ^. chanVoices

chanUserIsInvited :: ByteString -> IrcChannel -> Bool
chanUserIsInvited un ch = S.member un $ ch ^. chanInvites

chanUserMaySpeak :: ByteString -> IrcChannel -> Bool
chanUserMaySpeak un ch
  | (chanHasUser un ch || not (chanHasModeFlag NoOutsideMsgs ch))
     && not (chanHasModeFlag Moderated ch) = True
  | chanUserIsOper un ch                   = True
  | chanUserHasVoice un ch                 = True
  | otherwise                              = False

chanUserMayJoin :: ByteString -> IrcChannel -> Bool
chanUserMayJoin un ch | not $ chanHasModeFlag InviteOnly ch = True
                      | chanUserIsInvited un ch             = True
                      | otherwise                           = False

chanUserMaySetTopic :: ByteString -> IrcChannel -> Bool
chanUserMaySetTopic un ch
  | not (chanHasModeFlag TopicOperOnly ch) &&
    chanHasUser un ch    = True
  | chanUserIsOper un ch = True
  | otherwise            = False

-- | IrcServer management

ircAddUser :: ByteString -> IrcUser -> IrcServer -> IrcServer
ircAddUser nn usr = ircUsers %~ M.insert nn usr

ircDelUser :: ByteString -> IrcServer -> IrcServer
ircDelUser nn srv =
  srv & ircNicks %~ S.delete nn
      & if ircHasUser nn srv then let
           Just usr = M.lookup nn $ srv ^. ircUsers
           uchans = S.elems (usr ^. userChannels)
           ichans = S.elems (usr ^. userInvites)
          in
           (ircUsers %~ M.delete nn) .
           (ircChannels %~ \cs -> foldr (M.alter $ ircPartChan nn) cs uchans) .
           (ircChannels %~ \cs -> foldr (M.adjust $ chanDelInvite nn) cs ichans)
        else id

ircHasUser :: ByteString -> IrcServer -> Bool
ircHasUser nn srv = isJust $ M.lookup nn (srv ^. ircUsers)

ircHasChan :: ByteString -> IrcServer -> Bool
ircHasChan cn srv = isJust $ M.lookup cn (srv ^. ircChannels)

ircJoin :: ByteString -> ByteString -> IrcServer -> IrcServer
ircJoin un cn = (ircChannels %~ M.alter alterChan cn)
              . (ircUsers %~ M.adjust (userAddChan cn) un)
  where
    alterChan mChan = Just $ chanAddUser un (fromMaybe (newChannel un) mChan)

ircPartChan :: ByteString -> Maybe IrcChannel -> Maybe IrcChannel
ircPartChan un chan = case chanDelUser un (fromJust chan) of
                            IrcChannel{ _chanUsers = us }
                              | us == S.empty             -> Nothing
                            chan'                         -> Just chan'

ircPart :: ByteString -> ByteString -> IrcServer -> IrcServer
ircPart un cn srv =
  srv & (ircChannels %~ (M.alter $ ircPartChan un) cn)
      & (ircUsers %~ \us -> foldr (M.adjust $ userDelInvite cn) us iusers)
  where
    chan = fromJust $ M.lookup cn (srv ^. ircChannels)
    iusers = S.elems $ chan ^. chanInvites

ircInvite :: ByteString -> ByteString -> IrcServer -> IrcServer
ircInvite un cn = (ircChannels %~ M.adjust (chanAddInvite un) cn)
                  . (ircUsers %~ M.adjust (userAddInvite cn) un)

ircInviteCheck :: ByteString -> IrcChannel -> Bool
ircInviteCheck n chan =
  chanHasModeFlag InviteOnly chan && not (chanUserIsInvited n chan)

ircKeyCheck :: Maybe ByteString -> IrcChannel -> Bool
ircKeyCheck k chan =
  chanHasKey chan && (isNothing k || not (chanCheckKey (fromJust k) chan))

-- | Misc

parseParamList :: ByteString -> [ByteString]
parseParamList ps = filter (not . BS.null) $ BS.split ',' ps

zipParams :: [ByteString] -> [ByteString]
          -> [(ByteString, Maybe ByteString)]
zipParams chans chkeys = zip chans (map Just chkeys ++ repeat Nothing)

-- | Pretty Printing

ppServiceName :: ServiceName -> ByteString
ppServiceName = BS.pack

ppHostPreference :: HostPreference -> ByteString
ppHostPreference hp = case hp of
  HostAny  -> "*"
  HostIPv4 -> "*4"
  HostIPv6 -> "*6"
  Host hn  -> BS.pack hn

-- | Monadic utilities

useChan :: ByteString -> IrcMonad (Maybe IrcChannel)
useChan cname = fmap (M.lookup cname) $ use (clientServer . ircChannels)

useNick :: IrcMonad (Maybe ByteString)
useNick = do
  regState <- use clientReg
  return $ case regState of
    Unreg _ (Just nn) _       -> Just nn
    RegUser (NickName nn _ _) -> Just nn
    _                         -> Nothing

validateNick :: ByteString -> IrcMonad Bool
validateNick nickname = do
  nickSet <- use $ clientServer . ircNicks
  if S.member nickname nickSet
   then do
      tellNumeric err_nicknameinuse [nickname, ":Nickname is already in use."]
      return False
   else return True

channelTargets :: ByteString -> IrcMonad [Int]
channelTargets chname = do
  srv <- use clientServer
  Just mynick <- useNick
  let cUsers chan = S.elems (S.delete mynick $ chan ^. chanUsers)
  let chmap = srv ^. ircChannels
  case M.lookup chname chmap of
    Just chan -> fmap catMaybes $ forM (cUsers chan) userTarget
    _         -> do tellNumeric err_nosuchnick [chname, ":No such nick/channel"]
                    return []

userTarget :: ByteString -> IrcMonad (Maybe Int)
userTarget uname = do
  srv <- use clientServer
  let umap = srv ^. ircUsers
  case M.lookup uname umap of
    Just u -> return . Just $ u ^. userConn
    _      -> do tellNumeric err_nosuchnick [uname, ":No such nick/channel"]
                 return Nothing

findReceivers :: [ByteString] -> IrcMonad [([Int], ByteString)]
findReceivers rcvNames =
  fmap catMaybes $ forM rcvNames $ \name ->
    if BS.head name == '#'
     then do
      cids <- channelTargets name
      return $ case cids of
        [] -> Nothing
        cs -> Just (cs, name)
     else do
      cid <- userTarget name
      return $ case cid of
        Just c  -> Just ([c], name)
        Nothing -> Nothing

fwdMsg :: IrcMessage -> [([Int], ByteString)] -> IrcMonad ()
fwdMsg msg rcvs = forM_ rcvs $ \(cId, n) -> do
  -- replace multiple targets with the single target we're doing
  let m = msg{ params = n : tail (params msg) }
  tell [Msg m cId]

fwdMsgNoReplace :: IrcMessage -> [([Int], ByteString)] -> IrcMonad ()
fwdMsgNoReplace msg rcvs = do
  m <- addUserPrefix msg
  forM_ rcvs $ \(cId, _) -> tell [Msg m cId]


addUserPrefix :: IrcMessage -> IrcMonad IrcMessage
addUserPrefix msg = do
  regState <- use clientReg
  return $ case regState ^? regdNick of
    Just nickname -> msg{ prefix = Just . Right $ nickname }
    _             -> msg{ prefix = Nothing }

chanEcho :: [ByteString] -> IrcMessage -> IrcMonad ()
chanEcho chans iMsg = do
  msg <- addUserPrefix iMsg
  findReceivers chans >>= fwdMsg msg

allChanEcho :: IrcMessage -> IrcMonad ()
allChanEcho iMsg = do
  mNick <- useNick
  when (isJust mNick) $ do
    let nn = fromJust mNick
    mUser <- use $ clientServer . ircUsers . at nn
    when (isJust mUser) $ do
      let usr = fromJust mUser
      let chans = S.elems $ usr ^. userChannels
      chanEcho chans iMsg

-- | Adding responses to the Writer portion of the monad

tellYOURHOST :: ByteString -> IrcMonad ()
tellYOURHOST nickname = do
  srvname <- view ircHostName
  srvhost <- view ircHost
  srvport <- view ircPort
  version <- use $ clientServer . ircVersion
  let hostStr = ppHostPreference srvhost
      portStr = ppServiceName srvport
  tellNumeric rpl_yourhost
    [ nickname
    , BS.concat [ ":Your host is ", srvname
                , "[", hostStr, "/", portStr, "], "
                , "running version ", version ]
    ]

tellMOTD :: ByteString -> IrcMonad ()
tellMOTD nickname = do
  motd <- view ircMotd
  tellNumeric rpl_motdstart [nickname, ":- Message of the Day -"]
  forM_ motd $ \line ->
    tellNumeric rpl_motd [nickname, ":- " `BS.append` line]
  tellNumeric rpl_endofmotd [nickname, ":End of MOTD"]

tellWELCOME :: ByteString -> IrcMonad ()
tellWELCOME nickname = do
  srvname <- view ircHostName
  tellNumeric rpl_welcome [ nickname
                          , BS.append ":Welcome to IRC on " srvname ]

tellTOPIC :: ByteString -> IrcMonad ()
tellTOPIC cname = do
  chan <- use $ clientServer . ircChannels . at cname
  case chan of
    Just ch -> when (isJust $ ch ^. chanTopic) $
                 tellNumeric rpl_topic [cname, fromJust $ ch ^. chanTopic]
    Nothing -> return ()

tellNAMES :: [ByteString] -> IrcMonad ()
tellNAMES cnames = do
  Just nn <- useNick
  forM_ cnames $ \cname -> do
    mChan <- use $ clientServer . ircChannels . at cname
    case mChan of
      Just chan -> do
        let chanType = chanSigil chan
        nickL <- forM (S.elems $ chan ^. chanUsers) $ \uname ->
          return $ BS.append (chanUserSigil uname chan) uname
        let cUsers = BS.append ":" (BS.intercalate " " nickL)
        tellNumeric rpl_namreply [nn, chanType, cname, cUsers]
        tellNumeric rpl_endofnames [nn, head cnames, ":End of /NAMES list"]
      Nothing -> return ()

tellPrefixedCommand :: IrcCommand -> [IrcParam] -> IrcMonad ()
tellPrefixedCommand reply desc = do
  srvname <- view ircHostName
  connId <- use clientConn
  let msg = IrcMessage (Just $ Left srvname) (Left reply) desc
  tell [Msg msg [connId]]

tellCommand :: IrcCommand -> [IrcParam] -> IrcMonad ()
tellCommand reply desc = do
  connId <- use clientConn
  let msg = IrcMessage Nothing (Left reply) desc
  tell [Msg msg [connId]]

tellNumeric :: IrcReply -> [IrcParam] -> IrcMonad ()
tellNumeric reply desc = do
  srvname <- view ircHostName
  connId  <- use clientConn
  let msg = IrcMessage (Just $ Left srvname) (Right reply) desc
  tell [Msg msg [connId]]
  return ()
