{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards   #-}

module Pipes.IRC.Server.MessageHandler
  ( ircMessageHandler )
where

import           Control.Error
import           Control.Lens
import           Control.Monad.RWS
import qualified Data.ByteString.Char8     as BS
import           Data.Map                  as M
import           Data.Maybe
import           Data.Set                  as S
import           Pipes.IRC.Message.Types
import           Pipes.IRC.Server.IrcMonad
import           Pipes.IRC.Server.Types


ircMessageHandler :: IrcMessage -> IrcMonad ()
ircMessageHandler msg =
  -- drop messages that have prefixes (until we have Server links)
  when (isNothing $ prefix msg) $ do
    cReg <- use clientReg
    case cReg of
      Unreg   {} -> unregHandler msg
      RegUser {} -> regHandler   msg
    return ()

unregHandler :: IrcMessage -> IrcMonad ()
unregHandler msg@IrcMessage{..} =
  case command of
    Left PASS -> unregPASS msg
    Left NICK -> unregNICK msg
    Left USER -> unregUSER msg
    Left PONG -> handlePONG msg
    Left PING -> handlePING msg
    Left QUIT -> handleQUIT msg
    _         -> return ()

unregPASS :: IrcMessage -> IrcMonad ()
unregPASS IrcMessage{..} =
  if length params < 1
  then tellNumeric err_needmoreparams ["PASS", ":Need more parameters"]
  else do
    clientReg . rcvdPass .= (Just $ head params)
    return ()

unregNICK :: IrcMessage -> IrcMonad ()
unregNICK IrcMessage{..} =
  if length params /= 1
  then tellNumeric err_nonicknamegiven [":must supply a nickname"]
  else let nickname = head params in
    validateNick nickname >>= flip when
      (clientReg . rcvdNick .= Just nickname >>
       clientServer . ircNicks %= S.insert nickname >>
       tryRegistration)


unregUSER :: IrcMessage -> IrcMonad ()
unregUSER IrcMessage{..} =
  if length params < 4
  then tellNumeric err_needmoreparams ["USER", ":need more parameters"]
  else do
    clientReg . rcvdName .= (Just $ head params)
    tryRegistration

renderQuitMsg :: Maybe BS.ByteString -> IrcMonad BS.ByteString
renderQuitMsg (Just msg) = return $ BS.append "Quit: " msg
renderQuitMsg Nothing    = useNick >>= \nickname ->
  return $ case nickname of
    Just nick -> BS.append "Quit: " nick
    Nothing   -> ""

handlePING :: IrcMessage -> IrcMonad ()
handlePING _ = do
  srvname <- view ircHostName
  tellCommand PONG [":" <> srvname]

handlePONG :: IrcMessage -> IrcMonad ()
handlePONG _ = do
  cid <- use clientConn
  tell [Pong cid]

-- JOIN, MODE, KICK, PART, QUIT and of course PRIVMSG/NOTICE need to be
-- echoed to channels that the user belongs to

doQuit :: IrcMessage -> Maybe BS.ByteString -> IrcMonad ()
doQuit msg quitParamIn = do
  connId <- use clientConn
  quitMsg <- renderQuitMsg quitParamIn
  hostname <- use clientHost
  let hoststr = fromMaybe "unknown hostname" hostname
  let quitParam = BS.concat [ ":Closing Link: ", hoststr
                            , " (", quitMsg, ")"]
  tellCommand ERROR [quitParam]
  allChanEcho msg{params = [quitParam]}
  tell [Close connId]
  return ()

handleQUIT :: IrcMessage -> IrcMonad ()
handleQUIT msg@IrcMessage{..} =
  doQuit msg $ case params of
    []  -> Nothing
    p:_ -> Just p

handleJOIN :: IrcMessage -> IrcMonad ()
handleJOIN msg@IrcMessage{..} =
  case params of
    []      -> tellNumeric err_needmoreparams ["JOIN", ":Not enough parameters"]
    ["0"]   -> do
      Just nn <- useNick
      cs <- use $ clientServer . ircUsers . at nn
      let Just chans = fmap (^. userChannels) cs
      doPart msg{command=Left PART} (S.elems chans) Nothing
    cs:[]   -> doJoin msg $ zipParams (parseParamList cs) []
    cs:ks:_ -> doJoin msg $ zipParams (parseParamList cs) (parseParamList ks)

handlePART :: IrcMessage -> IrcMonad ()
handlePART msg@IrcMessage{..} =
  case params of
    []      -> tellNumeric err_needmoreparams ["PART", ":Not enough parameters"]
    cs:[]   -> doPart msg (parseParamList cs) Nothing
    cs:pm:_ -> doPart msg (parseParamList cs) (Just pm)

tellErr :: IrcReply -> [IrcParam] -> IrcMonadErr ()
tellErr r ps = lift (tellNumeric r ps) >> left []

checkRegistration :: IrcMonadErr BS.ByteString
checkRegistration = do
  mNick <- lift useNick
  when (isNothing mNick) $
    tellErr err_notregistered [":You have not registered"]
  right $ fromJust mNick

doJoin :: IrcMessage -> [(BS.ByteString, Maybe BS.ByteString)] -> IrcMonad ()
doJoin msg chans = forM_ chans $ \(c, k) -> runEitherT $ do
  nick <- checkRegistration
  mChan <- lift $ useChan c
  when (isJust mChan) $ do
    let chan = fromJust mChan

    when (chanHasUser nick chan) $ left "already on channel"

    when (ircInviteCheck nick chan) $
      tellErr err_inviteonlychan [c, ":Cannot join channel (+i)"]

    when (ircKeyCheck k chan) $
      tellErr err_badchannelkey [c, ":Cannot join channel (+k)"]

  lift $ do
    clientServer %= ircJoin nick c
    tellTOPIC c
    tellNAMES [c]
    chanEcho [c] msg

doPart :: IrcMessage -> [BS.ByteString] -> Maybe BS.ByteString -> IrcMonad ()
doPart msg chans pmsg = forM_ chans $ \c -> runEitherT $ do
  nn <- checkRegistration
  mChan <- lift . use $ clientServer . ircChannels . at c
  when (isNothing mChan) $ tellErr err_nosuchchannel [c, ":No such channel"]
  let ch = fromJust mChan
  unless (chanHasUser nn ch) $ tellErr err_notonchannel [c, ":Not on channel"]

  lift $ do
    let plist = case pmsg of
          Just bs -> [c, BS.append ":" bs]
          Nothing -> [c]
    let newMsg = msg{params = plist}
    chanEcho [c] newMsg
    findReceivers [nn] >>= fwdMsgNoReplace msg
    clientServer %= ircPart nn c

tryRegistration :: IrcMonad ()
tryRegistration = do
  regState <- use clientReg
  hostname <- use clientHost
  case regState of
    Unreg _ (Just nickname) (Just name) -> do
      let nn = NickName nickname (Just name) hostname
      clientReg .= RegUser nn
      usr <- mkUser
      clientServer . ircUsers %= M.insert nickname usr

      tellWELCOME  nickname
      tellYOURHOST nickname
      tellMOTD     nickname
    _                               -> return ()

regHandler :: IrcMessage -> IrcMonad ()
regHandler msg@IrcMessage{..} = do
  pMsg <- addUserPrefix msg
  case command of
    Left PRIVMSG -> handlePRIVMSG pMsg
    Left JOIN    -> handleJOIN pMsg
    Left PART    -> handlePART pMsg
    Left LIST    -> return ()
    Left NICK    -> return ()
    Left PING    -> handlePING pMsg
    Left PONG    -> handlePONG pMsg
    Left QUIT    -> handleQUIT pMsg
    _            -> return ()

handlePRIVMSG :: IrcMessage -> IrcMonad ()
handlePRIVMSG msg@IrcMessage{..} = do
  case params of
    []       -> tellNumeric err_norecipient []
    _:[]     -> tellNumeric err_notexttosend []
    rsp:_:_  -> let rs = parseParamList rsp
                in findReceivers rs >>= fwdMsg msg
  return ()
