{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE TemplateHaskell            #-}

module Pipes.IRC.Server.Types
  ( module Pipes.IRC.Server.Types
  , HostPreference (..), ServiceName, SockAddr, Socket
  )
where

import           Control.Concurrent.STM  (TVar)
import           Control.Error
import           Control.Lens
import           Control.Monad.RWS       (RWS)
import           Data.ByteString         (ByteString)
import           Data.Map                (Map)
import           Data.Set                (Set)
import           Data.Time.Clock         (UTCTime)
import           Pipes.Concurrent        (Output)
import           Pipes.IRC.Message.Types (IrcMessage, NickName)
import           Pipes.Network.TCP       (HostPreference (..), ServiceName,
                                          SockAddr, Socket)

type IrcEvents = [IrcEvent]

data IrcEvent = Msg { _outMsg  :: !IrcMessage
                    , _outDest :: ![Int]
                    }
              | Close { _closeConn :: Int }
              | Pong { _pongConn :: Int }
               deriving (Show)
makeLenses ''IrcEvent

data IrcUserMode = Away | Invisible | WallOps | Restricted
                 | Oper | LocalOper | ServerNotices
            deriving (Show, Eq, Enum, Ord)

data IrcUser =
  IrcUser { _userServerName :: !ByteString
          , _userModes      :: !(Set IrcUserMode)
          , _userChannels   :: !(Set ByteString)
          , _userConn       :: !Int
          , _userInvites    :: !(Set ByteString)
          } deriving (Show, Eq)
makeLenses ''IrcUser

data IrcChanModeFlags = Anonymous | InviteOnly | Moderated | NoOutsideMsgs
                      | Quiet | Private | Secret | TopicOperOnly
            deriving (Show, Eq, Enum, Ord)

data IrcChannel =
  IrcChannel { _chanTopic     :: !(Maybe ByteString)
             , _chanKey       :: !(Maybe ByteString)
             , _chanModeFlags :: !(Set IrcChanModeFlags)
             , _chanUsers     :: !(Set ByteString)
             , _chanOpers     :: !(Set ByteString)
             , _chanVoices    :: !(Set ByteString)
             , _chanInvites   :: !(Set ByteString)
             } deriving (Show, Eq)
makeLenses ''IrcChannel

data IrcServer =
  IrcServer { _ircNicks    :: !(Set ByteString)
            , _ircUsers    :: !(Map ByteString IrcUser)
            , _ircChannels :: !(Map ByteString IrcChannel)
            , _ircVersion  :: !ByteString
            } deriving (Show)
makeLenses ''IrcServer

data IrcConfig =
  IrcConfig { _ircPort     :: !ServiceName
            , _ircHost     :: !HostPreference
            , _ircHostName :: !ByteString
            , _ircMotd     :: ![ByteString]
            , _ircPass     :: !(Maybe ByteString)
            } deriving (Show)
makeLenses ''IrcConfig

data RegState = Unreg   { _rcvdPass :: !(Maybe ByteString)
                        , _rcvdNick :: !(Maybe ByteString)
                        , _rcvdName :: !(Maybe ByteString) }
              | RegUser { _regdNick :: !NickName }
  deriving (Show)
makeLenses ''RegState

data IrcConnection =
  IrcConnection { _sock    :: !Socket
                , _addr    :: !SockAddr
                , _hname   :: !(Maybe ByteString)
                , _out     :: !(Output IrcMessage)
                , _reg     :: !RegState
                , _lastCom :: !UTCTime
                , _gotPong :: !Bool
                }
makeLenses ''IrcConnection

data ServerState =
  ServerState { _ircState       :: !(TVar IrcServer)
              , _ircConfig      :: !IrcConfig
              , _ircConnections :: !(TVar (Map Int IrcConnection))
              , _ircConnIds     :: !(TVar Int)
              }
makeLenses ''ServerState

data ClientState =
  ClientState { _clientReg    :: !RegState
              , _clientServer :: !IrcServer
              , _clientHost   :: !(Maybe ByteString)
              , _clientConn   :: !Int
              } deriving (Show)
makeLenses ''ClientState

type IrcMonad a = RWS IrcConfig IrcEvents ClientState a
type IrcMonadErr a = EitherT String (RWS IrcConfig IrcEvents ClientState) a
